﻿using HarmonyLib;
using RimWorld;
using System;
using System.Collections.Generic;
using System.Reflection;
using System.Reflection.Emit;
using UnityEngine;
using Verse;

namespace Anthrosonae
{
    public class FurPattern
    {
        public Color first = Color.white;
        public Color second = Color.white;
        public float chance = 1f;
        public ShaderTypeDef shader;
    }
    public class GeneExtension : DefModExtension
    {
        public List<FurPattern> furPatterns;
        public bool applyFurPattern;
    }

    public class FurGene : Gene
    {
        public Color? first;
        public Color? second;
        public override void ExposeData()
        {
            base.ExposeData();
            Scribe_Values.Look(ref first, "first");
            Scribe_Values.Look(ref second, "second");
        }
    }


    public class AnthrosonaeMod : Mod
    {
        public AnthrosonaeMod(ModContentPack pack) : base(pack)
        {
            new Harmony("AnthrosonaeMod").PatchAll();
        }
    }

    [HarmonyPatch]
    public static class CharacterEditor_SetSkinColor_Patch
    {
        public static List<MethodBase> targetMethods;
        public static bool Prepare()
        {
            if (ModsConfig.IsActive("void.charactereditor"))
            {
                targetMethods = new List<MethodBase>();
                var assemblies = LoadedModManager.RunningModsListForReading.FirstOrDefault(x => x.PackageIdPlayerFacing 
                == "void.charactereditor").assemblies.loadedAssemblies;
                var field = AccessTools.Field(typeof(Pawn_StoryTracker), "skinColorOverride");
                foreach (var assembly in assemblies)
                {
                    try
                    {
                        foreach (var type in assembly.GetTypes())
                        {
                            try
                            {
                                var allMethods = type.GetMethods(AccessTools.all);
                                foreach (var method in allMethods)
                                {
                                    try
                                    {
                                        var parameters = method.GetParameters();
                                        if (parameters.Length == 3)
                                        {
                                            if (parameters[0].ParameterType == typeof(Pawn) && parameters[1].ParameterType == typeof(bool)
                                                && parameters[2].ParameterType == typeof(Color))
                                            {
                                                var instructions = PatchProcessor.GetCurrentInstructions(method);
                                                if (instructions.Any(x => x.opcode == OpCodes.Ldstr && x.OperandIs("skin") || 
                                                x.opcode == OpCodes.Stfld && x.OperandIs(field)))
                                                {
                                                    targetMethods.Add(method);
                                                }
                                            }
                                        }
                                    }
                                    catch { }
                                }
                            }
                            catch { }
                        }
                    }
                    catch { }
                }

                if (targetMethods.Any())
                {
                    return true;
                }
            }

            return false;
        }
        public static IEnumerable<MethodBase> TargetMethods()
        {
            return targetMethods;
        }

        public static void Postfix(Pawn __0, bool __1, Color __2)
        {
            if (__0.genes != null)
            {
                foreach (var gene in __0.genes.GenesListForReading)
                {
                    if (gene.Active && gene is FurGene furGene)
                    {
                        if (__1)
                        {
                            furGene.first = __2;
                        }
                        else
                        {
                            furGene.second = __2;
                        }
                    }
                }
            }
        }
    }

    [HarmonyPatch(typeof(GeneGraphicData), "GetGraphics")]
    public static class GeneGraphicData_GetGraphics_Patch
    {
        public static Dictionary<GeneGraphicData, Gene> mappedValues = new Dictionary<GeneGraphicData, Gene>();
        public static void Postfix(ref (Graphic, Graphic) __result, GeneGraphicData __instance)
        {
            var furPattern = PawnGraphicSet_ResolveAllGraphics_Patch.furPattern;
            if (furPattern != null && mappedValues.TryGetValue(__instance, out var gene) && gene.Active)
            {
                var extension = gene.def.GetModExtension<GeneExtension>();
                if (extension != null && extension.applyFurPattern) 
                {
                    __result.Item1 = __result.Item1.ApplyColors(furPattern.first, furPattern.second, furPattern.shader?.Shader);
                    __result.Item2 = __result.Item2.ApplyColors(furPattern.first, furPattern.second, furPattern.shader?.Shader);
                }
            }
        }
    }

    [HarmonyPatch(typeof(PawnGraphicSet), "ResolveAllGraphics")]
    public static class PawnGraphicSet_ResolveAllGraphics_Patch
    {
        public static FurPattern furPattern;
        public static void Postfix(PawnGraphicSet __instance)
        {
            furPattern = null;
            Pawn pawn = __instance.pawn;
            if (ModLister.BiotechInstalled && pawn.RaceProps.Humanlike && pawn.genes != null)
            {
                foreach (var gene in pawn.genes.GenesListForReading)
                {
                    if (gene.def.graphicData != null)
                    {
                        GeneGraphicData_GetGraphics_Patch.mappedValues[gene.def.graphicData] = gene;
                    }

                    if (gene.Active && gene is FurGene furGene)
                    {
                        var extension = gene.def.GetModExtension<GeneExtension>();
                        furPattern = new FurPattern();
                        if (extension?.furPatterns != null)
                        {
                            Rand.PushState();
                            Rand.Seed = pawn.thingIDNumber;
                            var furPatternTmp = extension.furPatterns.RandomElementByWeight(x => x.chance);
                            furPattern.first = furPatternTmp.first;
                            furPattern.second = furPatternTmp.second;
                            furPattern.shader = furPatternTmp.shader;
                            Rand.PopState();
                        }

                        if (furGene.first != null)
                            furPattern.first = furGene.first.Value;
                        if (furGene.second != null)
                            furPattern.second = furGene.second.Value;

                        __instance.furCoveredGraphic = GraphicDatabase.Get<Graphic_Multi>
                            (pawn.story.furDef.GetFurBodyGraphicPath(pawn),
                            furPattern.shader != null ? furPattern.shader.Shader : ShaderDatabase.CutoutSkinOverlay, Vector2.one,
                            furPattern.first, furPattern.second);
                        __instance.headGraphic = __instance.headGraphic.ApplyColors(furPattern.first, furPattern.second, furPattern.shader?.Shader);
                        __instance.pawn.story.skinColorOverride = furPattern.first;
                    }
                }

                __instance.CalculateHairMats();
                __instance.ResolveApparelGraphics();
                __instance.ResolveGeneGraphics();
            }

            furPattern = null;
        }

        public static Graphic ApplyColors(this Graphic graphic, Color first, Color second, Shader shaderOverride = null)
        {
            var graphicCopy = GraphicDatabase.Get(graphic.GetType(),
                graphic.path, shaderOverride ?? graphic.Shader, graphic.drawSize, first, second);
            return graphicCopy;
        }
    }
}
